import random
import numpy as np

from torch.utils import data
from torchvision import transforms
from .Datautils import Datautils
import albumentations as A

class BaseDataset(data.Dataset):

    def __init__(self, cfg, mode='train'):
        super(BaseDataset, self).__init__()

        self.utils = Datautils()
        self.mode = mode 
        
        img_list = cfg.img_list
        gt_list = cfg.gt_list
        train_val_split = cfg.train_val_split
        train_ratio = cfg.train_ratio

        self.retry = cfg.retry
        self.retry_mode = 'random'

        img_filename_list = self.utils.load_filelist(img_list)
        gt_filename_list = self.utils.load_filelist(gt_list)

        if train_val_split:
            train_list, val_list = self.utils.random_sublist(img_filename_list, num=int(train_ratio*len(img_filename_list)))
            gt_train, gt_val = self.utils.random_sublist(gt_filename_list, num=int(train_ratio*len(gt_filename_list)))
            if self.mode == 'train':
                img_filename_list = train_list
                gt_filename_list = gt_train 
            if self.mode == 'val':
                img_filename_list = val_list
                gt_filename_list = gt_val                
        
        self.imgs = img_filename_list
        self.gts = gt_filename_list
        mean_value = [0.11475467, 0.16522441, 0.14482306, 0.66098086]
        std_value = [0.03292060, 0.04573239, 0.05909370, 0.16499337]

        self.normalize = transforms.Normalize(mean=mean_value, std=std_value)
        self.transforms = transforms.Compose([
            transforms.ToTensor(),  
            self.normalize,
        ])

        self.augmentor = A.Compose([
            A.RandomSizedCrop(cfg.crop_size, cfg.crop_size),
            A.HorizontalFlip(),
            A.VerticalFlip(),
            A.RandomRotate90(),
            A.Transpose(),
            # A.OpticalDistortion(),  
        ])


    def __getitem__(self, index):
        retry = 0
        while retry < self.retry:
            try:
                img_filename = self.imgs[index]
                gt_filename = self.gts[index]
                img, meta = self.utils.read_raster(img_filename)
                gt, gt_meta = self.utils.read_raster(gt_filename)

                if meta['dtype'] == 'uint16':
                    img = (img / (2**12 - 1)).astype(np.float32)  # for planet imagery 

                img, gt = self.augmentor(image=img, mask=gt)
                img = self.transforms(img)
            except:
                retry += 1
                if self.retry_mode == 'random':
                    index = random.randit(0, len(self.imgs) - 1)
                else:
                    index += 1
        return img, gt

    def __len__(self):
        return len(self.imgs)
    
