import torch

import datasets
import models
from models.generate_model import generate_model
from configs import base_cfg
from optims import optims, scheduler
import criterions
from tqdm import tqdm


def train(**kwargs):
    base_cfg.parse(kwargs)

    model_params = generate_model(base_cfg)
    model = models.create_model(**model_params)
    model = model.cuda()

    # dataloader
    train_data = getattr(datasets, base_cfg.dataset)(base_cfg, 'train')
    train_dataloader = torch.utils.data.DataLoader(train_data, base_cfg.batch_size, shuffle=True, num_workers=base_cfg.num_workers, pin_memory=base_cfg.pin_memory, drop_last=True)
    val_data = getattr(datasets, base_cfg.dataset)(base_cfg, 'val')
    val_dataloader = torch.utils.data.DataLoader(val_data, base_cfg.batch_size, shuffle=True, num_workers=base_cfg.num_workers, pin_memory=base_cfg.pin_memory)
    
    # criterion 
    criterion = criterions.generate_criterion(base_cfg)

    # optimizer
    lr = base_cfg.lr
    optimizer = optims(model, base_cfg)
    schedule = scheduler(optimizer, base_cfg)
        
    # train process
    for epoch in range(base_cfg.max_epoch):
        print('INFO', f'Epoch {epoch+1} started')
        model.train()

        epoch_loss = 0.0
        epoch_iter = 0.0

        pbar = tqdm(total=len(train_dataloader))
        for idx, (img, gt) in enumerate(train_dataloader):
            pbar.update(1)

            if base_cfg.use_gpu:
                img = img.cuda()
                gt = gt.cuda()

            optimizer.zero_grad()
            score = model(img)
            loss = criterion(score, gt)
            loss.backward()

            optimizer.step()

            epoch_loss += loss.item()
            epoch_iter += 1

            # print training process
            if idx % base_cfg.print_freq == base_cfg.print_freq - 1:
                for param_group in optimizer.param_groups:
                    current_lr = param_group['lr']
                print('INFO', f'Epoch:{epoch+1}, Iter:{idx+1}, Loss:{loss.item()}, lr:{current_lr}')
        
        pbar.close()

        epoch_loss /= epoch_iter
        print('INFO', f'Epoch:{epoch+1}, Loss:{epoch_loss}')

        # validation during training 
        valloss = val(model, val_dataloader, criterion)
        
        if valloss < best_loss:
            best_loss = valloss
        best_loss_model = f'./checkpoints/test/best_loss_model.pth'
        print('INFO', f'Saveing Best Loss Model: {best_loss_model}')

        model_state_dict = dict()
        if base_cfg.use_gpu:
            for item, value in model.state_dict().items():
                module_item = '.'.join(item.split('.')[1:])
                model_state_dict[module_item] = value
        else:
            for item, value in model.state_dict().items():
                model_state_dict[item] = value
        torch.save({
            'epoch': epoch,
            'model': base_cfg.model,
            'backbone': base_cfg.encoder_name,
            'model_state_dict': model_state_dict,
            'bestloss': best_loss,
            'optimizer': base_cfg.optimizer,
            'optimizer_state_dict': optimizer.state_dict()
            }, best_loss_model)


def val(model, valdataset, criterion, epoch):
    model.eval()
    val_loss = 0.0
    num_data = 0.0

    pbar = tqdm(total=len(valdataset))
    for idx, (img, gt) in enumerate(valdataset):
        pbar.update(1)
        with torch.no_grad():
            if base_cfg.use_gpu:
                img = img.cuda()
                gt = gt.cuda()
            score = model(img)

            val_loss += criterion(score, gt).item()
            num_data += 1
    pbar.close()
    val_loss /= num_data

    model.train()
    print('INFO', f'Epoch Val Loss: {val_loss}, Epoch: {epoch}')
    return val_loss

